import logging
from typing import Literal

import numpy as np
import pynndescent
import scipy
from scipy.sparse import csr_matrix, issparse

logger = logging.getLogger(__name__)


def _compute_transitions(X: csr_matrix, density_normalize: bool = True):
    """Code from scanpy.

    https://github.com/scverse/scanpy/blob/2e98705347ea484c36caa9ba10de1987b09081bf/scanpy/neighbors/__init__.py#L899
    """
    # TODO(adamgayoso): Refactor this with Jax
    # density normalization as of Coifman et al. (2005)
    # ensures that kernel matrix is independent of sampling density
    if density_normalize:
        # q[i] is an estimate for the sampling density at point i
        # it's also the degree of the underlying graph
        q = np.asarray(X.sum(axis=0))
        if not issparse(X):
            Q = np.diag(1.0 / q)
        else:
            Q = scipy.sparse.spdiags(1.0 / q, 0, X.shape[0], X.shape[0])
        K = Q @ X @ Q
    else:
        K = X

    # z[i] is the square root of the row sum of K
    z = np.sqrt(np.asarray(K.sum(axis=0)))
    if not issparse(K):
        Z = np.diag(1.0 / z)
    else:
        Z = scipy.sparse.spdiags(1.0 / z, 0, K.shape[0], K.shape[0])
    transitions_sym = Z @ K @ Z

    return transitions_sym


def _compute_eigen(
    transitions_sym: csr_matrix,
    n_comps: int = 15,
    sort: Literal["decrease", "increase"] = "decrease",
):
    """Compute eigen decomposition of transition matrix.

    https://github.com/scverse/scanpy/blob/2e98705347ea484c36caa9ba10de1987b09081bf/scanpy/neighbors/__init__.py
    """
    # TODO(adamgayoso): Refactor this with Jax
    matrix = transitions_sym
    # compute the spectrum
    if n_comps == 0:
        evals, evecs = scipy.linalg.eigh(matrix)
    else:
        n_comps = min(matrix.shape[0] - 1, n_comps)
        # ncv = max(2 * n_comps + 1, int(np.sqrt(matrix.shape[0])))
        ncv = None
        which = "LM" if sort == "decrease" else "SM"
        # it pays off to increase the stability with a bit more precision
        matrix = matrix.astype(np.float64)

        evals, evecs = scipy.sparse.linalg.eigsh(matrix, k=n_comps, which=which, ncv=ncv)
        evals, evecs = evals.astype(np.float32), evecs.astype(np.float32)
    if sort == "decrease":
        evals = evals[::-1]
        evecs = evecs[:, ::-1]

    return evals, evecs


def _get_sparse_matrix_from_indices_distances_numpy(indices, distances, n_obs, n_neighbors):
    """Code from scanpy."""
    n_nonzero = n_obs * n_neighbors
    indptr = np.arange(0, n_nonzero + 1, n_neighbors)
    D = csr_matrix(
        (
            distances.copy().ravel(),  # copy the data, otherwise strange behavior here
            indices.copy().ravel(),
            indptr,
        ),
        shape=(n_obs, n_obs),
    )
    D.eliminate_zeros()
    D.sort_indices()
    return D


def diffusion_nn(X: csr_matrix, k: int, n_comps: int = 100):
    """Diffusion-based neighbors.

    This function generates a nearest neighbour list from a connectivities matrix.
    This allows us to select a consistent number of nearest neighbors across all methods.

    This differs from the original scIB implemenation by leveraging diffusion maps. Here we
    embed the data with diffusion maps in which euclidean distance represents well the diffusion
    distance. We then use pynndescent to find the nearest neighbours in this embedding space.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_cells) with non-zero values
        representing connectivities.
    k
        Number of nearest neighbours to select.
    n_comps
        Number of components for diffusion map

    Returns
    -------
    Neighbors graph
    """
    transitions = _compute_transitions(X)
    evals, evecs = _compute_eigen(transitions, n_comps=n_comps)
    evals += 1e-8  # Avoid division by zero
    # Multiscale such that the number of steps t gets "integrated out"
    embedding = evecs
    scaled_evals = np.array([e if e == 1 else e / (1 - e) for e in evals])
    embedding *= scaled_evals
    nn_obj = pynndescent.NNDescent(embedding, n_neighbors=k + 1)
    neigh_inds, neigh_distances = nn_obj.neighbor_graph
    # We purposely ignore the first neighbor as it is the cell itself
    # It gets added back inside the kbet internal function
    neigh_graph = _get_sparse_matrix_from_indices_distances_numpy(
        neigh_inds[:, 1:], neigh_distances[:, 1:], X.shape[0], k
    )
    return neigh_graph



from itertools import product

import pytest

import scib_metrics
from tests.utils.sampling import categorical_sample, normal_sample, poisson_sample

PCR_COMPARISON_PARAMS = list(product([100], [100], [False, True]))


@pytest.mark.parametrize("n_obs, n_vars, categorical", PCR_COMPARISON_PARAMS)
def test_pcr_comparison(n_obs, n_vars, categorical):
    X_pre = poisson_sample(n_obs, n_vars, seed=0)
    X_post = poisson_sample(n_obs, n_vars, seed=1)
    covariate = categorical_sample(n_obs, int(n_obs / 5)) if categorical else normal_sample(n_obs, seed=0)

    score = scib_metrics.pcr_comparison(X_pre, X_post, covariate, scale=True)
    assert score >= 0 and score <= 1



from typing import Optional

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax import jit

from .._types import NdArray
from ._pca import pca
from ._utils import one_hot


def principal_component_regression(
    X: NdArray,
    covariate: NdArray,
    categorical: bool = False,
    n_components: Optional[int] = None,
) -> float:
    """Principal component regression (PCR) :cite:p:`buttner2018`.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_features).
    covariate
        Array of shape (n_cells,) or (n_cells, 1) representing batch/covariate values.
    categorical
        If True, batch will be treated as categorical and one-hot encoded.
    n_components:
        Number of components to compute, passed into :func:`~scib_metrics.utils.pca`.
        If None, all components are used.

    Returns
    -------
    pcr: float
        Principal component regression using the first n_components principal components.
    """
    if len(X.shape) != 2:
        raise ValueError("Dimension mismatch: X must be 2-dimensional.")
    if X.shape[0] != covariate.shape[0]:
        raise ValueError("Dimension mismatch: X and batch must have the same number of samples.")
    if categorical:
        covariate = np.asarray(pd.Categorical(covariate).codes)
    else:
        covariate = np.asarray(covariate)

    covariate = one_hot(covariate) if categorical else covariate.reshape((covariate.shape[0], 1))

    pca_results = pca(X, n_components=n_components)

    # Center inputs for no intercept
    covariate = covariate - jnp.mean(covariate, axis=0)
    pcr = _pcr(pca_results.coordinates, covariate, pca_results.variance)
    return float(pcr)


@jit
def _pcr(
    X_pca: NdArray,
    covariate: NdArray,
    var: NdArray,
) -> NdArray:
    """Principal component regression.

    Parameters
    ----------
    X_pca
        Array of shape (n_cells, n_components) containing PCA coordinates. Must be standardized.
    covariate
        Array of shape (n_cells, 1) or (n_cells, n_classes) containing batch/covariate values. Must be standardized
        if not categorical (one-hot).
    var
        Array of shape (n_components,) containing the explained variance of each PC.
    """

    def r2(pc, batch):
        residual_sum = jnp.linalg.lstsq(batch, pc)[1]
        total_sum = jnp.sum((pc - jnp.mean(pc)) ** 2)
        return jnp.maximum(0, 1 - residual_sum / total_sum)

    # Index PCs on axis = 1, don't index batch
    r2_ = jax.vmap(r2, in_axes=(1, None))(X_pca, covariate)
    return jnp.dot(jnp.ravel(r2_), var) / jnp.sum(var)



from dataclasses import dataclass

import numpy as np


@dataclass
class NeighborsOutput:
    """Output of the nearest neighbors function.

    Attributes
    ----------
    distances : np.ndarray
        Array of distances to the nearest neighbors.
    indices : np.ndarray
        Array of indices of the nearest neighbors.
    """

    indices: np.ndarray
    distances: np.ndarray



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------
from typing import Any
import subprocess
import os
import importlib
import inspect
import re
import sys
from datetime import datetime
from importlib.metadata import metadata
from pathlib import Path

HERE = Path(__file__).parent
sys.path.insert(0, str(HERE / "extensions"))


# -- Project information -----------------------------------------------------

project_name = "scib-metrics"
info = metadata(project_name)
package_name = "scib_metrics"
author = info["Author"]
copyright = f"{datetime.now():%Y}, {author}."
version = info["Version"]

# The full version, including alpha/beta/rc tags
release = info["Version"]

bibtex_bibfiles = ["references.bib"]
templates_path = ["_templates"]
nitpicky = True  # Warn about broken links
needs_sphinx = "4.0"

html_context = {
    "display_github": True,  # Integrate GitHub
    "github_user": "yoseflab",  # Username
    "github_repo": project_name,  # Repo name
    "github_version": "main",  # Version
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings.
# They can be extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "myst_nb",
    "sphinx.ext.autodoc",
    "sphinx_copybutton",
    "sphinx.ext.linkcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinxcontrib.bibtex",
    "sphinx_autodoc_typehints",
    "sphinx.ext.mathjax",
    "IPython.sphinxext.ipython_console_highlighting",
    *[p.stem for p in (HERE / "extensions").glob("*.py")],
]

autosummary_generate = True
autodoc_member_order = "groupwise"
default_role = "literal"
bibtex_reference_style = "author_year"
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = True  # having a separate entry generally helps readability
napoleon_use_param = True
myst_heading_anchors = 3  # create anchors for h1-h3
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
    "html_admonition",
]
myst_url_schemes = ("http", "https", "mailto")
nb_output_stderr = "remove"
nb_execution_mode = "off"
nb_merge_streams = True
typehints_defaults = "braces"

source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".myst": "myst-nb",
}

intersphinx_mapping = {
    "anndata": ("https://anndata.readthedocs.io/en/stable/", None),
    "ipython": ("https://ipython.readthedocs.io/en/stable/", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "python": ("https://docs.python.org/3", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "scanpy": ("https://scanpy.readthedocs.io/en/stable/", None),
    "jax": ("https://jax.readthedocs.io/en/latest/", None),
    "plottable": ("https://plottable.readthedocs.io/en/latest/", None),
}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# -- Linkcode settings -------------------------------------------------


def git(*args):
    """Run a git command and return the output."""
    return subprocess.check_output(["git", *args]).strip().decode()


# https://github.com/DisnakeDev/disnake/blob/7853da70b13fcd2978c39c0b7efa59b34d298186/docs/conf.py#L192
# Current git reference. Uses branch/tag name if found, otherwise uses commit hash
git_ref = None
try:
    git_ref = git("name-rev", "--name-only", "--no-undefined", "HEAD")
    git_ref = re.sub(r"^(remotes/[^/]+|tags)/", "", git_ref)
except Exception:  # noqa: B902
    pass

# (if no name found or relative ref, use commit hash instead)
if not git_ref or re.search(r"[\^~]", git_ref):
    try:
        git_ref = git("rev-parse", "HEAD")
    except Exception:  # noqa: B902
        git_ref = "main"

# https://github.com/DisnakeDev/disnake/blob/7853da70b13fcd2978c39c0b7efa59b34d298186/docs/conf.py#L192
github_repo = "https://github.com/" + html_context["github_user"] + "/" + project_name
_project_module_path = os.path.dirname(importlib.util.find_spec(package_name).origin)  # type: ignore


def linkcode_resolve(domain, info):
    """Resolve links for the linkcode extension."""
    if domain != "py":
        return None

    try:
        obj: Any = sys.modules[info["module"]]
        for part in info["fullname"].split("."):
            obj = getattr(obj, part)
        obj = inspect.unwrap(obj)

        if isinstance(obj, property):
            obj = inspect.unwrap(obj.fget)  # type: ignore

        path = os.path.relpath(inspect.getsourcefile(obj), start=_project_module_path)  # type: ignore
        src, lineno = inspect.getsourcelines(obj)
    except Exception:  # noqa: B902
        return None

    path = f"{path}#L{lineno}-L{lineno + len(src) - 1}"
    return f"{github_repo}/blob/{git_ref}/src/{package_name}/{path}"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_title = "scib-metrics"

html_theme_options = {
    "repository_url": github_repo,
    "use_repository_button": True,
}

pygments_style = "default"

nitpick_ignore = [
    # If building the documentation fails because of a missing link that is outside your control,
    # you can add an exception to this list.
]


def setup(app):
    """App setup hook."""
    app.add_config_value(
        "recommonmark_config",
        {
            "auto_toc_tree_section": "Contents",
            "enable_auto_toc_tree": True,
            "enable_math": True,
            "enable_inline_math": False,
            "enable_eval_rst": True,
        },
        True,
    )



import warnings
from typing import Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as np
from chex import ArrayDevice
from jax import nn
from scipy.sparse import csr_matrix
from sklearn.neighbors import NearestNeighbors
from sklearn.utils import check_array

from .._types import ArrayLike, IntOrKey, NdArray


def get_ndarray(x: ArrayDevice) -> np.ndarray:
    """Convert Jax device array to Numpy array."""
    return np.array(jax.device_get(x))


def one_hot(y: NdArray, n_classes: Optional[int] = None) -> jnp.ndarray:
    """One-hot encode an array. Wrapper around :func:`~jax.nn.one_hot`.

    Parameters
    ----------
    y
        Array of shape (n_cells,) or (n_cells, 1).
    n_classes
        Number of classes. If None, inferred from the data.

    Returns
    -------
    one_hot: jnp.ndarray
        Array of shape (n_cells, n_classes).
    """
    n_classes = n_classes or jnp.max(y) + 1
    return nn.one_hot(jnp.ravel(y), n_classes)


def validate_seed(seed: IntOrKey) -> jax.random.KeyArray:
    """Validate a seed and return a Jax random key."""
    return jax.random.PRNGKey(seed) if isinstance(seed, int) else seed


def check_square(X: ArrayLike):
    """Check if a matrix is square."""
    if X.shape[0] != X.shape[1]:
        raise ValueError("X must be a square matrix")


def convert_knn_graph_to_idx(X: csr_matrix) -> Tuple[np.ndarray, np.ndarray]:
    """Convert a kNN graph to indices and distances."""
    check_array(X, accept_sparse="csr")
    check_square(X)

    n_neighbors = np.unique(X.nonzero()[0], return_counts=True)[1]
    if len(np.unique(n_neighbors)) > 1:
        raise ValueError("Each cell must have the same number of neighbors.")

    n_neighbors = int(np.unique(n_neighbors)[0])
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Precomputed sparse input")
        nn_obj = NearestNeighbors(n_neighbors=n_neighbors, metric="precomputed").fit(X)
        kneighbors = nn_obj.kneighbors(X)
    return kneighbors



from ._diffusion_nn import diffusion_nn
from ._dist import cdist, pdist_squareform
from ._kmeans import KMeansJax
from ._lisi import compute_simpson_index
from ._pca import pca
from ._pcr import principal_component_regression
from ._silhouette import silhouette_samples
from ._utils import check_square, convert_knn_graph_to_idx, get_ndarray, one_hot

__all__ = [
    "silhouette_samples",
    "cdist",
    "pdist_squareform",
    "get_ndarray",
    "KMeansJax",
    "pca",
    "principal_component_regression",
    "one_hot",
    "compute_simpson_index",
    "convert_knn_graph_to_idx",
    "check_square",
    "diffusion_nn",
]



import logging
from importlib.metadata import version

from . import nearest_neighbors, utils
from ._graph_connectivity import graph_connectivity
from ._isolated_labels import isolated_labels
from ._kbet import kbet, kbet_per_label
from ._lisi import clisi_knn, ilisi_knn, lisi_knn
from ._nmi_ari import nmi_ari_cluster_labels_kmeans, nmi_ari_cluster_labels_leiden
from ._pcr_comparison import pcr_comparison
from ._settings import settings
from ._silhouette import silhouette_batch, silhouette_label

__all__ = [
    "utils",
    "nearest_neighbors",
    "isolated_labels",
    "pcr_comparison",
    "silhouette_label",
    "silhouette_batch",
    "ilisi_knn",
    "clisi_knn",
    "lisi_knn",
    "nmi_ari_cluster_labels_kmeans",
    "nmi_ari_cluster_labels_leiden",
    "kbet",
    "kbet_per_label",
    "graph_connectivity",
]

__version__ = version("scib-metrics")

settings.verbosity = logging.INFO
# Jax sets the root logger, this prevents double output.
logger = logging.getLogger("scib_metrics")
logger.propagate = False



import numpy as np
from pynndescent import NNDescent

from ._dataclass import NeighborsOutput


def pynndescent(X: np.ndarray, n_neighbors: int, random_state: int = 0, n_jobs: int = 1) -> NeighborsOutput:
    """Run pynndescent approximate nearest neighbor search.

    Parameters
    ----------
    X
        Data matrix.
    n_neighbors
        Number of neighbors to search for.
    random_state
        Random state.
    n_jobs
        Number of jobs to use.
    """
    # Variables from umap (https://github.com/lmcinnes/umap/blob/3f19ce19584de4cf99e3d0ae779ba13a57472cd9/umap/umap_.py#LL326-L327)
    # which is used by scanpy under the hood
    n_trees = min(64, 5 + int(round((X.shape[0]) ** 0.5 / 20.0)))
    n_iters = max(5, int(round(np.log2(X.shape[0]))))
    max_candidates = 60

    knn_search_index = NNDescent(
        X,
        n_neighbors=n_neighbors,
        random_state=random_state,
        low_memory=True,
        n_jobs=n_jobs,
        compressed=False,
        n_trees=n_trees,
        n_iters=n_iters,
        max_candidates=max_candidates,
    )
    indices, distances = knn_search_index.neighbor_graph

    return NeighborsOutput(indices=indices, distances=distances)



import anndata
import numpy as np
from scipy.sparse import csr_matrix
from sklearn.neighbors import NearestNeighbors

import scib_metrics


def dummy_x_labels(return_symmetric_positive=False, x_is_neighbors_graph=False):
    np.random.seed(1)
    X = np.random.normal(size=(100, 10))
    labels = np.random.randint(0, 2, size=(100,))
    if return_symmetric_positive:
        X = np.abs(X @ X.T)
    if x_is_neighbors_graph:
        dist_mat = csr_matrix(scib_metrics.utils.cdist(X, X))
        nbrs = NearestNeighbors(n_neighbors=30, algorithm="kd_tree").fit(X)
        X = nbrs.kneighbors_graph(X)
        X = X.multiply(dist_mat)
    return X, labels


def dummy_x_labels_batch(x_is_neighbors_graph=False):
    X, labels = dummy_x_labels(x_is_neighbors_graph=x_is_neighbors_graph)
    batch = np.random.randint(0, 2, size=(100,))
    return X, labels, batch


def dummy_benchmarker_adata():
    X, labels, batch = dummy_x_labels_batch(x_is_neighbors_graph=False)
    adata = anndata.AnnData(X)
    labels_key = "labels"
    batch_key = "batch"
    adata.obs[labels_key] = labels
    adata.obs[batch_key] = batch
    embedding_keys = []
    for i in range(5):
        key = f"X_emb_{i}"
        adata.obsm[key] = X
        embedding_keys.append(key)
    return adata, embedding_keys, labels_key, batch_key



import warnings

from ._types import NdArray
from .utils import principal_component_regression


def pcr_comparison(
    X_pre: NdArray,
    X_post: NdArray,
    covariate: NdArray,
    scale: bool = True,
    **kwargs,
) -> float:
    """Principal component regression (PCR) comparison :cite:p:`buttner2018`.

    Compare the explained variance before and after integration.

    Parameters
    ----------
    X_pre
        Pre-integration array of shape (n_cells, n_features).
    X_post
        Post-integration array of shape (n_celss, n_features).
    covariate_pre:
        Array of shape (n_cells,) or (n_cells, 1) representing batch/covariate values.
    scale
        Whether to scale the score between 0 and 1. If True, larger values correspond to
        larger differences in variance contributions between `X_pre` and `X_post`.
    kwargs
        Keyword arguments passed into :func:`~scib_metrics.principal_component_regression`.

    Returns
    -------
    pcr_compared: float
        Principal component regression score comparing the explained variance before and
        after integration.
    """
    if X_pre.shape[0] != X_post.shape[0]:
        raise ValueError("Dimension mismatch: `X_pre` and `X_post` must have the same number of samples.")
    if covariate.shape[0] != X_pre.shape[0]:
        raise ValueError("Dimension mismatch: `X_pre` and `covariate` must have the same number of samples.")

    pcr_pre = principal_component_regression(X_pre, covariate, **kwargs)
    pcr_post = principal_component_regression(X_post, covariate, **kwargs)

    if scale:
        pcr_compared = (pcr_pre - pcr_post) / pcr_pre
        if pcr_compared < 0:
            warnings.warn(
                "PCR comparison score is negative, meaning variance contribution "
                "increased after integration. Setting to 0."
            )
            pcr_compared = 0
    else:
        pcr_compared = pcr_post - pcr_pre

    return pcr_compared



from itertools import product

import numpy as np
import pandas as pd
import pytest
from scib.metrics import pc_regression

import scib_metrics
from scib_metrics.utils import get_ndarray
from tests.utils.sampling import categorical_sample, normal_sample, poisson_sample

PCR_PARAMS = list(product([10, 100, 1000], [10, 100, 1000], [False]))
# TODO(martinkim0): Currently not testing categorical covariates because of
# TODO(martinkim0): reproducibility issues with original scib. See comment in PR #16.


@pytest.mark.parametrize("n_obs, n_vars, categorical", PCR_PARAMS)
def test_pcr(n_obs, n_vars, categorical):
    def _test_pcr(n_obs: int, n_vars: int, n_components: int, categorical: bool, eps=1e-3, seed=123):
        X = poisson_sample(n_obs, n_vars, seed=seed)
        covariate = categorical_sample(n_obs, int(n_obs / 5)) if categorical else normal_sample(n_obs, seed=seed)

        pcr_true = pc_regression(
            get_ndarray(X),
            pd.Categorical(get_ndarray(covariate)) if categorical else get_ndarray(covariate),
            n_comps=n_components,
        )
        pcr = scib_metrics.utils.principal_component_regression(
            X,
            covariate,
            categorical=categorical,
            n_components=n_components,
        )
        assert np.allclose(pcr_true, pcr, atol=eps)

    max_components = min(n_obs, n_vars)
    _test_pcr(n_obs, n_vars, n_components=max_components, categorical=categorical)
    _test_pcr(n_obs, n_vars, n_components=max_components - 1, categorical=categorical)
    _test_pcr(n_obs, n_vars, n_components=int(max_components / 2), categorical=categorical)
    _test_pcr(n_obs, n_vars, n_components=1, categorical=categorical)



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix

from scib_metrics.utils import compute_simpson_index, convert_knn_graph_to_idx


def lisi_knn(X: csr_matrix, labels: np.ndarray, perplexity: float = None) -> np.ndarray:
    """Compute the local inverse simpson index (LISI) for each cell :cite:p:`korsunsky2019harmony`.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_cells) with non-zero values
        representing distances to exactly each cell's k nearest neighbors.
    labels
        Array of shape (n_cells,) representing label values
        for each cell.
    perplexity
        Parameter controlling effective neighborhood size. If None, the
        perplexity is set to the number of neighbors // 3.

    Returns
    -------
    lisi
        Array of shape (n_cells,) with the LISI score for each cell.
    """
    labels = np.asarray(pd.Categorical(labels).codes)
    knn_dists, knn_idx = convert_knn_graph_to_idx(X)

    if perplexity is None:
        perplexity = np.floor(knn_idx.shape[1] / 3)

    n_labels = len(np.unique(labels))

    simpson = compute_simpson_index(knn_dists, knn_idx, labels, n_labels, perplexity=perplexity)
    return 1 / simpson


def ilisi_knn(X: csr_matrix, batches: np.ndarray, perplexity: float = None, scale: bool = True) -> np.ndarray:
    """Compute the integration local inverse simpson index (iLISI) for each cell :cite:p:`korsunsky2019harmony`.

    Returns a scaled version of the iLISI score for each cell, by default :cite:p:`luecken2022benchmarking`.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_cells) with non-zero values
        representing distances to exactly each cell's k nearest neighbors.
    batches
        Array of shape (n_cells,) representing batch values
        for each cell.
    perplexity
        Parameter controlling effective neighborhood size. If None, the
        perplexity is set to the number of neighbors // 3.
    scale
        Scale lisi into the range [0, 1]. If True, higher values are better.

    Returns
    -------
    ilisi
        Array of shape (n_cells,) with the iLISI score for each cell.
    """
    batches = np.asarray(pd.Categorical(batches).codes)
    lisi = lisi_knn(X, batches, perplexity=perplexity)
    ilisi = np.nanmedian(lisi)
    if scale:
        nbatches = len(np.unique(batches))
        ilisi = (ilisi - 1) / (nbatches - 1)
    return ilisi


def clisi_knn(X: csr_matrix, labels: np.ndarray, perplexity: float = None, scale: bool = True) -> np.ndarray:
    """Compute the cell-type local inverse simpson index (cLISI) for each cell :cite:p:`korsunsky2019harmony`.

    Returns a scaled version of the cLISI score for each cell, by default :cite:p:`luecken2022benchmarking`.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_cells) with non-zero values
        representing distances to exactly each cell's k nearest neighbors.
    labels
        Array of shape (n_cells,) representing cell type label values
        for each cell.
    perplexity
        Parameter controlling effective neighborhood size. If None, the
        perplexity is set to the number of neighbors // 3.
    scale
        Scale lisi into the range [0, 1]. If True, higher values are better.

    Returns
    -------
    clisi
        Array of shape (n_cells,) with the cLISI score for each cell.
    """
    labels = np.asarray(pd.Categorical(labels).codes)
    lisi = lisi_knn(X, labels, perplexity=perplexity)
    clisi = np.nanmedian(lisi)
    if scale:
        nlabels = len(np.unique(labels))
        clisi = (nlabels - clisi) / (nlabels - 1)
    return clisi



from functools import partial
from typing import Literal

import jax
import jax.numpy as jnp
import numpy as np
from sklearn.utils import check_array

from .._types import IntOrKey
from ._dist import cdist, pdist_squareform
from ._utils import get_ndarray, validate_seed


def _initialize_random(X: jnp.ndarray, n_clusters: int, pdists: jnp.ndarray, key: jax.random.KeyArray) -> jnp.ndarray:
    """Initialize cluster centroids randomly."""
    n_obs = X.shape[0]
    indices = jax.random.choice(key, n_obs, (n_clusters,), replace=False)
    initial_state = X[indices]
    return initial_state


@partial(jax.jit, static_argnums=1)
def _initialize_plus_plus(
    X: jnp.ndarray, n_clusters: int, pdists: jnp.ndarray, key: jax.random.KeyArray
) -> jnp.ndarray:
    """Initialize cluster centroids with k-means++ algorithm."""

    def _init(key, pdists):
        key, subkey = jax.random.split(key)
        n_obs = pdists.shape[0]
        # sample first centroid uniformly at random
        idx = jax.random.choice(subkey, n_obs)
        centroids = jnp.full((n_clusters,), -1, dtype=jnp.int32).at[0].set(idx)
        mask = jnp.zeros((n_obs,), dtype=jnp.bool_).at[idx].set(True)
        return centroids, mask, pdists, key

    def _step(state):
        centroids, mask, pdists, key = state
        key, subkey = jax.random.split(key)
        n_obs = pdists.shape[0]
        # d(x) = min_{mu in centroids} ||x - mu||^2, d(x) = 0 if x in centroids
        probs = jnp.where(mask, 0, jnp.min(jnp.where(mask, pdists, jnp.inf), axis=1) ** 2)
        # sample with probability ~ d(x)
        idx = jax.random.choice(subkey, n_obs, p=probs / jnp.sum(probs))
        centroids = centroids.at[jnp.sum(mask)].set(idx)
        mask = mask.at[idx].set(True)
        return centroids, mask, pdists, key

    def _convergence(state):
        _, mask, _, _ = state
        return jnp.sum(mask) < n_clusters

    centroids, _, _, _ = jax.lax.while_loop(_convergence, _step, _init(key, pdists))
    return X[centroids]


@jax.jit
def _get_dist_labels(X: jnp.ndarray, centroids: jnp.ndarray) -> jnp.ndarray:
    """Get the distance and labels for each observation."""
    dist = cdist(X, centroids)
    labels = jnp.argmin(dist, axis=1)
    return dist, labels


class KMeansJax:
    """Jax implementation of :class:`sklearn.cluster.KMeans`.

    This implementation is limited to Euclidean distance.

    Parameters
    ----------
    n_clusters
        Number of clusters.
    init
        Cluster centroid initialization method. One of the following:

        * ``'k-means++'``: Sample initial cluster centroids based on an
            empirical distribution of the points' contributions to the
            overall inertia.
        * ``'random'``: Uniformly sample observations as initial centroids
    n_init
        Number of times the k-means algorithm will be initialized.
    max_iter
        Maximum number of iterations of the k-means algorithm for a single run.
    tol
        Relative tolerance with regards to inertia to declare convergence.
    seed
        Random seed.
    """

    def __init__(
        self,
        n_clusters: int = 8,
        init: Literal["k-means++", "random"] = "random",
        n_init: int = 10,
        max_iter: int = 300,
        tol: float = 1e-4,
        seed: IntOrKey = 0,
    ):
        self.n_clusters = n_clusters
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.seed: jax.random.KeyArray = validate_seed(seed)

        if init not in ["k-means++", "random"]:
            raise ValueError("Invalid init method, must be one of ['k-means++' or 'random'].")
        if init == "k-means++":
            self._initialize = _initialize_plus_plus
        else:
            self._initialize = _initialize_random

    def fit(self, X: np.ndarray):
        """Fit the model to the data."""
        X = check_array(X, dtype=np.float32, order="C")
        # Subtract mean for numerical accuracy
        mean = X.mean(axis=0)
        X -= mean
        self._fit(X)
        X += mean
        self.cluster_centroids_ += mean
        return self

    def _fit(self, X: np.ndarray):
        self._pdists = pdist_squareform(X)
        all_centroids, all_inertias = jax.lax.map(
            lambda key: self._kmeans_full_run(X, key), jax.random.split(self.seed, self.n_init)
        )
        i = jnp.argmin(all_inertias)
        self.cluster_centroids_ = get_ndarray(all_centroids[i])
        self.inertia_ = get_ndarray(all_inertias[i])
        _, labels = _get_dist_labels(X, self.cluster_centroids_)
        self.labels_ = get_ndarray(labels)
        del self._pdists

    @partial(jax.jit, static_argnums=(0,))
    def _kmeans_full_run(self, X: jnp.ndarray, key: jnp.ndarray) -> jnp.ndarray:
        def _kmeans_step(state):
            old_inertia = state[1]
            centroids, _, _, n_iter = state
            # TODO(adamgayoso): Efficiently compute argmin and min simultaneously.
            dist, new_labels = _get_dist_labels(X, centroids)
            # From https://colab.research.google.com/drive/1AwS4haUx6swF82w3nXr6QKhajdF8aSvA?usp=sharing
            counts = (new_labels[jnp.newaxis, :] == jnp.arange(self.n_clusters)[:, jnp.newaxis]).sum(
                axis=1, keepdims=True
            )
            counts = jnp.clip(counts, a_min=1, a_max=None)
            # Sum over points in a centroid by zeroing others out
            new_centroids = (
                jnp.sum(
                    jnp.where(
                        # axes: (data points, clusters, data dimension)
                        new_labels[:, jnp.newaxis, jnp.newaxis]
                        == jnp.arange(self.n_clusters)[jnp.newaxis, :, jnp.newaxis],
                        X[:, jnp.newaxis, :],
                        0.0,
                    ),
                    axis=0,
                )
                / counts
            )
            new_inertia = jnp.mean(jnp.min(dist, axis=1))
            n_iter = n_iter + 1
            return new_centroids, new_inertia, old_inertia, n_iter

        def _kmeans_convergence(state):
            _, new_inertia, old_inertia, n_iter = state
            cond1 = jnp.abs(old_inertia - new_inertia) < self.tol
            cond2 = n_iter > self.max_iter
            return jnp.logical_or(cond1, cond2)[0]

        centroids = self._initialize(X, self.n_clusters, self._pdists, key)
        # centroids, new_inertia, old_inertia, n_iter
        state = (centroids, jnp.inf, jnp.inf, jnp.array([0.0]))
        state = _kmeans_step(state)
        state = jax.lax.while_loop(_kmeans_convergence, _kmeans_step, state)
        return state[0], state[1]



from typing import Union

import jax
import jax.numpy as jnp
import numpy as np
import scipy.sparse as sp

NdArray = Union[np.ndarray, jnp.ndarray]
IntOrKey = Union[int, jax.random.KeyArray]
ArrayLike = Union[np.ndarray, sp.spmatrix, jnp.ndarray]



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components


def graph_connectivity(X: csr_matrix, labels: np.ndarray) -> float:
    """Quantify the connectivity of the subgraph per cell type label.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_cells) with non-zero values
        representing distances to exactly each cell's k nearest neighbors.
    labels
        Array of shape (n_cells,) representing label values
        for each cell.
    """
    # TODO(adamgayoso): Utils for validating inputs
    clust_res = []

    for label in np.unique(labels):
        mask = labels == label
        graph_sub = X[mask]
        graph_sub = graph_sub[:, mask]
        _, comps = connected_components(graph_sub, connection="strong")
        tab = pd.value_counts(comps)
        clust_res.append(tab.max() / sum(tab))

    return np.mean(clust_res)



from typing import Optional, Tuple

import jax.numpy as jnp
from chex import dataclass
from jax import jit

from .._types import NdArray
from ._utils import get_ndarray


@dataclass
class _SVDResult:
    """SVD result.

    Attributes
    ----------
    u
        Array of shape (n_cells, n_components) containing the left singular vectors.
    s
        Array of shape (n_components,) containing the singular values.
    v
        Array of shape (n_components, n_features) containing the right singular vectors.
    """

    u: NdArray
    s: NdArray
    v: NdArray


@dataclass
class _PCAResult:
    """PCA result.

    Attributes
    ----------
    coordinates
        Array of shape (n_cells, n_components) containing the PCA coordinates.
    components
        Array of shape (n_components, n_features) containing the PCA components.
    variance
        Array of shape (n_components,) containing the explained variance of each PC.
    variance_ratio
        Array of shape (n_components,) containing the explained variance ratio of each PC.
    svd
        Dataclass containing the SVD data.
    """

    coordinates: NdArray
    components: NdArray
    variance: NdArray
    variance_ratio: NdArray
    svd: Optional[_SVDResult] = None


def _svd_flip(
    u: NdArray,
    v: NdArray,
    u_based_decision: bool = True,
):
    """Sign correction to ensure deterministic output from SVD.

    Jax implementation of :func:`~sklearn.utils.extmath.svd_flip`.

    Parameters
    ----------
    u
        Left singular vectors of shape (M, K).
    v
        Right singular vectors of shape (K, N).
    u_based_decision
        If True, use the columns of u as the basis for sign flipping.
    """
    if u_based_decision:
        max_abs_cols = jnp.argmax(jnp.abs(u), axis=0)
        signs = jnp.sign(u[max_abs_cols, jnp.arange(u.shape[1])])
    else:
        max_abs_rows = jnp.argmax(jnp.abs(v), axis=1)
        signs = jnp.sign(v[jnp.arange(v.shape[0]), max_abs_rows])
    u_ = u * signs
    v_ = v * signs[:, None]
    return u_, v_


def pca(
    X: NdArray,
    n_components: Optional[int] = None,
    return_svd: bool = False,
) -> _PCAResult:
    """Principal component analysis (PCA).

    Parameters
    ----------
    X
        Array of shape (n_cells, n_features).
    n_components
        Number of components to keep. If None, all components are kept.
    return_svd
        If True, also return the results from SVD.

    Returns
    -------
    results: _PCAData
    """
    max_components = min(X.shape)
    if n_components and n_components > max_components:
        raise ValueError(f"n_components = {n_components} must be <= min(n_cells, n_features) = {max_components}")
    n_components = n_components or max_components

    u, s, v, variance, variance_ratio = _pca(X)

    # Select n_components
    coordinates = u[:, :n_components] * s[:n_components]
    components = v[:n_components]
    variance_ = variance[:n_components]
    variance_ratio_ = variance_ratio[:n_components]

    results = _PCAResult(
        coordinates=get_ndarray(coordinates),
        components=get_ndarray(components),
        variance=get_ndarray(variance_),
        variance_ratio=get_ndarray(variance_ratio_),
        svd=_SVDResult(u=get_ndarray(u), s=get_ndarray(s), v=get_ndarray(v)) if return_svd else None,
    )
    return results


@jit
def _pca(
    X: NdArray,
) -> Tuple[NdArray, NdArray, NdArray, NdArray, NdArray]:
    """Principal component analysis.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_features).

    Returns
    -------
    u: NdArray
        Left singular vectors of shape (M, K).
    s: NdArray
        Singular values of shape (K,).
    v: NdArray
        Right singular vectors of shape (K, N).
    variance: NdArray
        Array of shape (K,) containing the explained variance of each PC.
    variance_ratio: NdArray
        Array of shape (K,) containing the explained variance ratio of each PC.
    """
    X_ = X - jnp.mean(X, axis=0)
    u, s, v = jnp.linalg.svd(X_, full_matrices=False)
    u, v = _svd_flip(u, v)

    variance = (s**2) / (X.shape[0] - 1)
    total_variance = jnp.sum(variance)
    variance_ratio = variance / total_variance

    return u, s, v, variance, variance_ratio



import logging
from functools import partial
from typing import Tuple, Union

import chex
import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
import scipy
from scipy.sparse import csr_matrix

from scib_metrics.utils import convert_knn_graph_to_idx, diffusion_nn, get_ndarray

from ._types import NdArray

logger = logging.getLogger(__name__)


def _chi2_cdf(df: Union[int, NdArray], x: NdArray) -> float:
    """Chi2 cdf.

    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.chdtr.html
    for explanation of gammainc.
    """
    return jax.scipy.special.gammainc(df / 2, x / 2)


@partial(jax.jit, static_argnums=2)
def _kbet(neigh_batch_ids: jnp.ndarray, batches: jnp.ndarray, n_batches: int) -> float:
    expected_freq = jnp.bincount(batches, length=n_batches)
    expected_freq = expected_freq / jnp.sum(expected_freq)
    dof = n_batches - 1

    observed_counts = jax.vmap(partial(jnp.bincount, length=n_batches))(neigh_batch_ids)
    expected_counts = expected_freq * neigh_batch_ids.shape[1]
    test_statistics = jnp.sum(jnp.square(observed_counts - expected_counts) / expected_counts, axis=1)
    p_values = 1 - jax.vmap(_chi2_cdf, in_axes=(None, 0))(dof, test_statistics)

    return test_statistics, p_values


def kbet(X: csr_matrix, batches: np.ndarray, alpha: float = 0.05) -> float:
    """Compute kbet :cite:p:`buttner2018`.

    This implementation is inspired by the implementation in Pegasus:
    https://pegasus.readthedocs.io/en/stable/index.html

    A higher acceptance rate means more mixing of batches. This implementation does
    not exactly mirror the default original implementation, as there is currently no
    `adapt` option.

    Note that this is also not equivalent to the kbet used in the original scib package,
    as that one computes kbet for each cell type label. To achieve this, use
    :func:`scib_metrics.kbet_per_label`.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_cells) with non-zero values
        representing distances to exactly each cell's k nearest neighbors.
    batches
        Array of shape (n_cells,) representing batch values
        for each cell.
    alpha
        Significance level for the statistical test.

    Returns
    -------
    acceptance_rate
        Kbet acceptance rate of the sample.
    stat_mean
        Mean Kbet chi-square statistic over all cells.
    pvalue_mean
        Mean Kbet p-value over all cells.

    Notes
    -----
    This function requires X to be cell-cell distances, not connectivies.
    """
    if len(batches) != X.shape[0]:
        raise ValueError("Length of batches does not match number of cells.")
    _, knn_idx = convert_knn_graph_to_idx(X)
    # Make sure self is included
    knn_idx = np.concatenate([np.arange(knn_idx.shape[0])[:, None], knn_idx], axis=1)
    batches = np.asarray(pd.Categorical(batches).codes)
    neigh_batch_ids = batches[knn_idx]
    chex.assert_equal_shape([neigh_batch_ids, knn_idx])
    n_batches = jnp.unique(batches).shape[0]
    test_statistics, p_values = _kbet(neigh_batch_ids, batches, n_batches)
    test_statistics = get_ndarray(test_statistics)
    p_values = get_ndarray(p_values)
    acceptance_rate = (p_values >= alpha).mean()

    return acceptance_rate, test_statistics, p_values


def kbet_per_label(
    X: csr_matrix,
    batches: np.ndarray,
    labels: np.ndarray,
    alpha: float = 0.05,
    diffusion_n_comps: int = 100,
    return_df: bool = False,
) -> Union[float, Tuple[float, pd.DataFrame]]:
    """Compute kBET score per cell type label as in :cite:p:`luecken2022benchmarking`.

    This approximates the method used in the original scib package. Notably, the underlying
    kbet might have some inconsistencies with the R implementation. Furthermore, to equalize
    the neighbor graphs of cell type subsets we use diffusion distance approximated with diffusion
    maps. Increasing `diffusion_n_comps` will increase the accuracy of the approximation.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_cells) with non-zero values
        representing connectivies to exactly each cell's k nearest neighbors.
    batches
        Array of shape (n_cells,) representing batch values
        for each cell.
    alpha
        Significance level for the statistical test.
    diffusion_n_comps
        Number of diffusion components to use for diffusion distance approximation.
    return_df
        Return dataframe of results in addition to score.

    Returns
    -------
    kbet_score
        Kbet score over all cells. Higher means more integrated, as in the kBET acceptance rate.
    df
        Dataframe with kBET score per cell type label.

    Notes
    -----
    This function requires X to be cell-cell connectivities, not distances.
    """
    if len(batches) != X.shape[0]:
        raise ValueError("Length of batches does not match number of cells.")
    if len(labels) != X.shape[0]:
        raise ValueError("Length of labels does not match number of cells.")
    # set upper bound for k0
    size_max = 2**31 - 1
    batches = np.asarray(pd.Categorical(batches).codes)
    labels = np.asarray(labels)

    # prepare call of kBET per cluster
    kbet_scores = {"cluster": [], "kBET": []}
    for clus in np.unique(labels):
        # subset by label
        mask = labels == clus
        X_sub = X[mask, :][:, mask]
        X_sub.sort_indices()
        n_obs = X_sub.shape[0]
        batches_sub = batches[mask]

        # check if neighborhood size too small or only one batch in subset
        if np.logical_or(n_obs < 10, len(np.unique(batches_sub)) == 1):
            logger.info(f"{clus} consists of a single batch or is too small. Skip.")
            score = np.nan
        else:
            quarter_mean = np.floor(np.mean(pd.Series(batches_sub).value_counts()) / 4).astype("int")
            k0 = np.min([70, np.max([10, quarter_mean])])
            # check k0 for reasonability
            if k0 * n_obs >= size_max:
                k0 = np.floor(size_max / n_obs).astype("int")

            n_comp, labs = scipy.sparse.csgraph.connected_components(X_sub, connection="strong")

            if n_comp == 1:  # a single component to compute kBET on
                try:
                    diffusion_n_comps = np.min([diffusion_n_comps, n_obs - 1])
                    nn_graph_sub = diffusion_nn(X_sub, k=k0, n_comps=diffusion_n_comps).astype("float")
                    # call kBET
                    score, _, _ = kbet(
                        nn_graph_sub,
                        batches=batches_sub,
                        alpha=alpha,
                    )
                except ValueError:
                    logger.info("Diffusion distance failed. Skip.")
                    score = 0  # i.e. 100% rejection

            else:
                # check the number of components where kBET can be computed upon
                comp_size = pd.value_counts(labs)
                # check which components are small
                comp_size_thresh = 3 * k0
                idx_nonan = np.flatnonzero(np.in1d(labs, comp_size[comp_size >= comp_size_thresh].index))

                # check if 75% of all cells can be used for kBET run
                if len(idx_nonan) / len(labs) >= 0.75:
                    # create another subset of components, assume they are not visited in a diffusion process
                    X_sub_sub = X_sub[idx_nonan, :][:, idx_nonan]
                    X_sub_sub.sort_indices()

                    try:
                        diffusion_n_comps = np.min([diffusion_n_comps, X_sub_sub.shape[0] - 1])
                        nn_graph_sub_sub = diffusion_nn(X_sub_sub, k=k0, n_comps=diffusion_n_comps).astype("float")
                        # call kBET
                        score, _, _ = kbet(
                            nn_graph_sub_sub,
                            batches=batches_sub[idx_nonan],
                            alpha=alpha,
                        )
                    except ValueError:
                        logger.info("Diffusion distance failed. Skip.")
                        score = 0  # i.e. 100% rejection
                else:  # if there are too many too small connected components, set kBET score to 0
                    score = 0  # i.e. 100% rejection

        kbet_scores["cluster"].append(clus)
        kbet_scores["kBET"].append(score)

    kbet_scores = pd.DataFrame.from_dict(kbet_scores)
    kbet_scores = kbet_scores.reset_index(drop=True)

    final_score = np.nanmean(kbet_scores["kBET"])
    if not return_df:
        return final_score
    else:
        return final_score, kbet_scores



from typing import Union

import jax
import jax.numpy as jnp

IntOrKey = Union[int, jax.random.KeyArray]


def _validate_seed(seed: IntOrKey) -> jax.random.KeyArray:
    return jax.random.PRNGKey(seed) if isinstance(seed, int) else seed


def categorical_sample(
    n_obs: int,
    n_cats: int,
    seed: IntOrKey = 0,
) -> jnp.ndarray:
    return jax.random.categorical(_validate_seed(seed), jnp.ones(n_cats), shape=(n_obs,))


def normal_sample(
    n_obs: int,
    mean: float = 0.0,
    var: float = 1.0,
    seed: IntOrKey = 0,
) -> jnp.ndarray:
    return jax.random.multivariate_normal(_validate_seed(seed), jnp.ones(n_obs) * mean, jnp.eye(n_obs) * var)


def poisson_sample(
    n_obs: int,
    n_vars: int,
    rate: float = 1.0,
    seed: IntOrKey = 0,
) -> jnp.ndarray:
    return jax.random.poisson(_validate_seed(seed), rate, shape=(n_obs, n_vars))



import jax
import jax.numpy as jnp
import numpy as np


@jax.jit
def _euclidean_distance(x: np.array, y: np.array) -> float:
    dist = jnp.sqrt(jnp.sum((x - y) ** 2))
    return dist


@jax.jit
def cdist(x: np.ndarray, y: np.ndarray) -> jnp.ndarray:
    """Jax implementation of :func:`scipy.spatial.distance.cdist`.

    Uses euclidean distance.

    Parameters
    ----------
    x
        Array of shape (n_cells_a, n_features)
    y
        Array of shape (n_cells_b, n_features)

    Returns
    -------
    dist
        Array of shape (n_cells_a, n_cells_b)
    """
    return jax.vmap(lambda x1: jax.vmap(lambda y1: _euclidean_distance(x1, y1))(y))(x)


@jax.jit
def pdist_squareform(X: np.ndarray) -> jnp.ndarray:
    """Jax implementation of :func:`scipy.spatial.distance.pdist` and :func:`scipy.spatial.distance.squareform`.

    Uses euclidean distance.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_features)

    Returns
    -------
    dist
        Array of shape (n_cells, n_cells)
    """
    n_cells = X.shape[0]
    inds = jnp.triu_indices(n_cells)

    def _body_fn(X, i_j):
        i, j = i_j
        return X, _euclidean_distance(X[i], X[j])

    dist_mat = jnp.zeros((n_cells, n_cells))
    dist_mat = dist_mat.at[inds].set(jax.lax.scan(_body_fn, X, (inds[0], inds[1]))[1])
    dist_mat = jnp.maximum(dist_mat, dist_mat.T)
    return dist_mat



from scib_metrics.nearest_neighbors import jax_approx_min_k
from tests.utils.data import dummy_benchmarker_adata


def test_jax_neighbors():
    ad, emb_keys, _, _ = dummy_benchmarker_adata()
    output = jax_approx_min_k(ad.obsm[emb_keys[0]], 10)
    assert output.distances.shape == (ad.n_obs, 10)



from itertools import product

import jax.numpy as jnp
import numpy as np
import pytest
from sklearn.decomposition import PCA

import scib_metrics
from tests.utils.sampling import poisson_sample

PCA_PARAMS = list(product([10, 100, 1000], [10, 100, 1000]))


@pytest.mark.parametrize("n_obs, n_vars", PCA_PARAMS)
def test_pca(n_obs: int, n_vars: int):
    def _test_pca(n_obs: int, n_vars: int, n_components: int, eps: float = 1e-4):
        X = poisson_sample(n_obs, n_vars)
        max_components = min(X.shape)
        pca = scib_metrics.utils.pca(X, n_components=n_components, return_svd=True)

        # SANITY CHECKS
        assert pca.coordinates.shape == (X.shape[0], n_components)
        assert pca.components.shape == (n_components, X.shape[1])
        assert pca.variance.shape == (n_components,)
        assert pca.variance_ratio.shape == (n_components,)
        # SVD should not be truncated to n_components
        assert pca.svd is not None
        assert pca.svd.u.shape == (X.shape[0], max_components)
        assert pca.svd.s.shape == (max_components,)
        assert pca.svd.v.shape == (max_components, X.shape[1])

        # VALUE CHECKS
        # TODO(martinkim0): Currently not checking coordinates and components,
        # TODO(martinkim0): implementations differ very slightly and not sure why.
        pca_true = PCA(n_components=n_components, svd_solver="full").fit(X)
        # assert np.allclose(pca_true.transform(X), pca.coordinates, atol=eps)
        # assert np.allclose(pca_true.components_, pca.components, atol=eps)
        assert np.allclose(pca_true.singular_values_, pca.svd.s[:n_components], atol=eps)
        assert np.allclose(pca_true.explained_variance_, pca.variance, atol=eps)
        assert np.allclose(pca_true.explained_variance_ratio_, pca.variance_ratio, atol=eps)
        # Use arpack iff n_components < max_components
        if n_components < max_components:
            pca_true = PCA(n_components=n_components, svd_solver="arpack").fit(X)
            # assert jnp.allclose(pca_true.transform(X), pca.coordinates, atol=eps)
            # assert jnp.allclose(pca_true.components_, pca.components, atol=eps)
            assert jnp.allclose(pca_true.singular_values_, pca.svd.s[:n_components], atol=eps)
            assert jnp.allclose(pca_true.explained_variance_, pca.variance, atol=eps)
            assert jnp.allclose(pca_true.explained_variance_ratio_, pca.variance_ratio, atol=eps)

    max_components = min(n_obs, n_vars)
    _test_pca(n_obs, n_vars, n_components=max_components)
    _test_pca(n_obs, n_vars, n_components=max_components - 1)
    _test_pca(n_obs, n_vars, n_components=int(max_components / 2))
    _test_pca(n_obs, n_vars, n_components=1)



import logging
from typing import Dict, Tuple

import numpy as np
import scanpy as sc
from scipy.sparse import spmatrix
from sklearn.metrics.cluster import adjusted_rand_score, normalized_mutual_info_score
from sklearn.utils import check_array

from .utils import KMeansJax, check_square

logger = logging.getLogger(__name__)


def _compute_clustering_kmeans(X: np.ndarray, n_clusters: int) -> np.ndarray:
    kmeans = KMeansJax(n_clusters)
    kmeans.fit(X)
    return kmeans.labels_


def _compute_clustering_leiden(connectivity_graph: spmatrix, resolution: float) -> np.ndarray:
    g = sc._utils.get_igraph_from_adjacency(connectivity_graph)
    clustering = g.community_leiden(objective_function="modularity", weights="weight", resolution_parameter=resolution)
    clusters = clustering.membership
    return np.asarray(clusters)


def _compute_nmi_ari_cluster_labels(
    X: np.ndarray,
    labels: np.ndarray,
    resolution: float = 1.0,
) -> Tuple[float, float]:
    labels_pred = _compute_clustering_leiden(X, resolution)
    nmi = normalized_mutual_info_score(labels, labels_pred, average_method="arithmetic")
    ari = adjusted_rand_score(labels, labels_pred)
    return nmi, ari


def nmi_ari_cluster_labels_kmeans(X: np.ndarray, labels: np.ndarray) -> Dict[str, float]:
    """Compute nmi and ari between k-means clusters and labels.

    This deviates from the original implementation in scib by using k-means
    with k equal to the known number of cell types/labels. This leads to
    a more efficient computation of the nmi and ari scores.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_features).
    labels
        Array of shape (n_cells,) representing label values

    Returns
    -------
    nmi
        Normalized mutual information score
    ari
        Adjusted rand index score
    """
    X = check_array(X, accept_sparse=False, ensure_2d=True)
    n_clusters = len(np.unique(labels))
    labels_pred = _compute_clustering_kmeans(X, n_clusters)
    nmi = normalized_mutual_info_score(labels, labels_pred, average_method="arithmetic")
    ari = adjusted_rand_score(labels, labels_pred)

    return {"nmi": nmi, "ari": ari}


def nmi_ari_cluster_labels_leiden(
    X: spmatrix, labels: np.ndarray, optimize_resolution: bool = True, resolution: float = 1.0, n_jobs: int = 1
) -> Dict[str, float]:
    """Compute nmi and ari between leiden clusters and labels.

    This deviates from the original implementation in scib by using leiden instead of
    louvain clustering. Installing joblib allows for parallelization of the leiden
    resoution optimization.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_cells) representing a connectivity graph.
        Values should represent weights between pairs of neighbors, with a higher weight
        indicating more connected.
    labels
        Array of shape (n_cells,) representing label values
    optimize_resolution
        Whether to optimize the resolution parameter of leiden clustering by searching over
        10 values
    resolution
        Resolution parameter of leiden clustering. Only used if optimize_resolution is False.
    n_jobs
        Number of jobs for parallelizing resolution optimization via joblib. If -1, all CPUs
        are used.

    Returns
    -------
    nmi
        Normalized mutual information score
    ari
        Adjusted rand index score
    """
    X = check_array(X, accept_sparse=True, ensure_2d=True)
    check_square(X)
    if optimize_resolution:
        n = 10
        resolutions = np.array([2 * x / n for x in range(1, n + 1)])
        try:
            from joblib import Parallel, delayed

            out = Parallel(n_jobs=n_jobs)(delayed(_compute_nmi_ari_cluster_labels)(X, labels, r) for r in resolutions)
        except ImportError:
            logger.info("Using for loop over resolutions. pip install joblib for parallelization.")
            out = [_compute_nmi_ari_cluster_labels(X, labels, r) for r in resolutions]
        nmi_ari = np.array(out)
        nmi_ind = np.argmax(nmi_ari[:, 0])
        nmi, ari = nmi_ari[nmi_ind, :]
        return {"nmi": nmi, "ari": ari}
    else:
        nmi, ari = _compute_nmi_ari_cluster_labels(X, labels, resolution)

    return {"nmi": nmi, "ari": ari}



import jax.numpy as jnp
import numpy as np
import pandas as pd
from harmonypy import compute_lisi as harmonypy_lisi
from scipy.sparse import csr_matrix
from scipy.spatial.distance import cdist as sp_cdist
from scipy.spatial.distance import pdist, squareform
from sklearn.metrics import silhouette_samples as sk_silhouette_samples
from sklearn.neighbors import NearestNeighbors

import scib_metrics
from tests.utils.data import dummy_x_labels, dummy_x_labels_batch

scib_metrics.settings.jax_fix_no_kernel_image()


def test_package_has_version():
    scib_metrics.__version__


def test_cdist():
    x = jnp.array([[1, 2], [3, 4]])
    y = jnp.array([[5, 6], [7, 8]])
    assert np.allclose(scib_metrics.utils.cdist(x, y), sp_cdist(x, y))


def test_pdist():
    x = jnp.array([[1, 2], [3, 4]])
    assert np.allclose(scib_metrics.utils.pdist_squareform(x), squareform(pdist(x)))


def test_silhouette_samples():
    X, labels = dummy_x_labels()
    assert np.allclose(scib_metrics.utils.silhouette_samples(X, labels), sk_silhouette_samples(X, labels), atol=1e-5)


def test_silhouette_label():
    X, labels = dummy_x_labels()
    score = scib_metrics.silhouette_label(X, labels)
    assert score > 0
    scib_metrics.silhouette_label(X, labels, rescale=False)


def test_silhouette_batch():
    X, labels, batch = dummy_x_labels_batch()
    score = scib_metrics.silhouette_batch(X, labels, batch)
    assert score > 0
    scib_metrics.silhouette_batch(X, labels, batch)


def test_compute_simpson_index():
    X, labels = dummy_x_labels()
    D = scib_metrics.utils.cdist(X, X)
    nbrs = NearestNeighbors(n_neighbors=30, algorithm="kd_tree").fit(X)
    D, knn_idx = nbrs.kneighbors(X)
    scib_metrics.utils.compute_simpson_index(
        jnp.array(D), jnp.array(knn_idx), jnp.array(labels), len(np.unique(labels))
    )


def test_lisi_knn():
    X, labels = dummy_x_labels()
    dist_mat = csr_matrix(scib_metrics.utils.cdist(X, X))
    nbrs = NearestNeighbors(n_neighbors=30, algorithm="kd_tree").fit(X)
    knn_graph = nbrs.kneighbors_graph(X)
    knn_graph = knn_graph.multiply(dist_mat)
    lisi_res = scib_metrics.lisi_knn(knn_graph, labels, perplexity=10)
    harmonypy_lisi_res = harmonypy_lisi(
        X, pd.DataFrame(labels, columns=["labels"]), label_colnames=["labels"], perplexity=10
    )[:, 0]
    assert np.allclose(lisi_res, harmonypy_lisi_res)


def test_ilisi_clisi_knn():
    X, labels, batches = dummy_x_labels_batch(x_is_neighbors_graph=True)
    scib_metrics.ilisi_knn(X, batches, perplexity=10)
    scib_metrics.clisi_knn(X, labels, perplexity=10)


def test_nmi_ari_cluster_labels_kmeans():
    X, labels = dummy_x_labels()
    out = scib_metrics.nmi_ari_cluster_labels_kmeans(X, labels)
    nmi, ari = out["nmi"], out["ari"]
    assert isinstance(nmi, float)
    assert isinstance(ari, float)


def test_nmi_ari_cluster_labels_leiden_parallel():
    X, labels = dummy_x_labels(return_symmetric_positive=True)
    out = scib_metrics.nmi_ari_cluster_labels_leiden(X, labels, optimize_resolution=True, n_jobs=2)
    nmi, ari = out["nmi"], out["ari"]
    assert isinstance(nmi, float)
    assert isinstance(ari, float)


def test_nmi_ari_cluster_labels_leiden_single_resolution():
    X, labels = dummy_x_labels(return_symmetric_positive=True)
    out = scib_metrics.nmi_ari_cluster_labels_leiden(X, labels, optimize_resolution=False, resolution=0.1)
    nmi, ari = out["nmi"], out["ari"]
    assert isinstance(nmi, float)
    assert isinstance(ari, float)


def test_isolated_labels():
    X, labels, batch = dummy_x_labels_batch()
    scib_metrics.isolated_labels(X, labels, batch)


def test_kmeans():
    X, _ = dummy_x_labels()
    kmeans = scib_metrics.utils.KMeansJax(2)
    kmeans.fit(X)
    assert kmeans.labels_.shape == (X.shape[0],)


def test_kbet():
    X, _, batch = dummy_x_labels_batch(x_is_neighbors_graph=True)
    acc_rate, stats, pvalues = scib_metrics.kbet(X, batch)
    assert isinstance(acc_rate, float)
    assert len(stats) == X.shape[0]
    assert len(pvalues) == X.shape[0]


def test_kbet_per_label():
    X, labels, batch = dummy_x_labels_batch(x_is_neighbors_graph=True)
    score = scib_metrics.kbet_per_label(X, batch, labels)
    assert isinstance(score, float)


def test_graph_connectivity():
    X, labels = dummy_x_labels(return_symmetric_positive=True)
    metric = scib_metrics.graph_connectivity(X, labels)
    assert isinstance(metric, float)



import pandas as pd

from scib_metrics.benchmark import BatchCorrection, Benchmarker, BioConservation
from scib_metrics.nearest_neighbors import jax_approx_min_k
from tests.utils.data import dummy_benchmarker_adata


def test_benchmarker():
    ad, emb_keys, batch_key, labels_key = dummy_benchmarker_adata()
    bm = Benchmarker(ad, batch_key, labels_key, emb_keys)
    bm.benchmark()
    results = bm.get_results()
    assert isinstance(results, pd.DataFrame)
    bm.plot_results_table()


def test_benchmarker_custom_metric_booleans():
    bioc = BioConservation(
        isolated_labels=False, nmi_ari_cluster_labels_leiden=False, silhouette_label=False, clisi_knn=True
    )
    bc = BatchCorrection(kbet_per_label=False, graph_connectivity=False, ilisi_knn=True)
    ad, emb_keys, batch_key, labels_key = dummy_benchmarker_adata()
    bm = Benchmarker(ad, batch_key, labels_key, emb_keys, batch_correction_metrics=bc, bio_conservation_metrics=bioc)
    bm.benchmark()
    results = bm.get_results(clean_names=False)
    assert isinstance(results, pd.DataFrame)
    assert "isolated_labels" not in results.columns
    assert "nmi_ari_cluster_labels_leiden" not in results.columns
    assert "silhouette_label" not in results.columns
    assert "clisi_knn" in results.columns
    assert "kbet_per_label" not in results.columns
    assert "graph_connectivity" not in results.columns
    assert "ilisi_knn" in results.columns


def test_benchmarker_custom_metric_callable():
    bioc = BioConservation(clisi_knn={"perplexity": 10})
    ad, emb_keys, batch_key, labels_key = dummy_benchmarker_adata()
    bm = Benchmarker(ad, batch_key, labels_key, emb_keys, bio_conservation_metrics=bioc)
    bm.benchmark()
    results = bm.get_results(clean_names=False)
    assert "clisi_knn" in results.columns


def test_benchmarker_custom_near_neighs():
    ad, emb_keys, batch_key, labels_key = dummy_benchmarker_adata()
    bm = Benchmarker(ad, batch_key, labels_key, emb_keys)
    bm.prepare(neighbor_computer=jax_approx_min_k)
    bm.benchmark()
    results = bm.get_results()
    assert isinstance(results, pd.DataFrame)
    bm.plot_results_table()



from functools import partial
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd

from ._dist import cdist
from ._utils import get_ndarray


@jax.jit
def _silhouette_reduce(
    D_chunk: jnp.ndarray, start: int, labels: jnp.ndarray, label_freqs: jnp.ndarray
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Accumulate silhouette statistics for vertical chunk of X.

    Follows scikit-learn implementation.

    Parameters
    ----------
    D_chunk
        Array of shape (n_chunk_samples, n_samples)
        Precomputed distances for a chunk.
    start
        First index in the chunk.
    labels
        Array of shape (n_samples,)
        Corresponding cluster labels, encoded as {0, ..., n_clusters-1}.
    label_freqs
        Distribution of cluster labels in ``labels``.
    """
    # accumulate distances from each sample to each cluster
    D_chunk_len = D_chunk.shape[0]

    # If running into memory issues, use fori_loop instead of vmap
    # clust_dists = jnp.zeros((D_chunk_len, len(label_freqs)), dtype=D_chunk.dtype)
    # def _bincount(i, _data):
    #     clust_dists, D_chunk, labels, label_freqs = _data
    #     clust_dists = clust_dists.at[i].set(jnp.bincount(labels, weights=D_chunk[i], length=label_freqs.shape[0]))
    #     return clust_dists, D_chunk, labels, label_freqs

    # clust_dists = jax.lax.fori_loop(
    #     0, D_chunk_len, lambda i, _data: _bincount(i, _data), (clust_dists, D_chunk, labels, label_freqs)
    # )[0]

    clust_dists = jax.vmap(partial(jnp.bincount, length=label_freqs.shape[0]), in_axes=(None, 0))(labels, D_chunk)

    # intra_index selects intra-cluster distances within clust_dists
    intra_index = (jnp.arange(D_chunk_len), jax.lax.dynamic_slice(labels, (start,), (D_chunk_len,)))
    # intra_clust_dists are averaged over cluster size outside this function
    intra_clust_dists = clust_dists[intra_index]
    # of the remaining distances we normalise and extract the minimum
    clust_dists = clust_dists.at[intra_index].set(jnp.inf)
    clust_dists /= label_freqs
    inter_clust_dists = clust_dists.min(axis=1)
    return intra_clust_dists, inter_clust_dists


def _pairwise_distances_chunked(X: jnp.ndarray, chunk_size: int, reduce_fn: callable) -> jnp.ndarray:
    """Compute pairwise distances in chunks to reduce memory usage."""
    n_samples = X.shape[0]
    n_chunks = jnp.ceil(n_samples / chunk_size).astype(int)
    intra_dists_all = []
    inter_dists_all = []
    for i in range(n_chunks):
        start = i * chunk_size
        end = min((i + 1) * chunk_size, n_samples)
        intra_cluster_dists, inter_cluster_dists = reduce_fn(cdist(X[start:end], X), start=start)
        intra_dists_all.append(intra_cluster_dists)
        inter_dists_all.append(inter_cluster_dists)
    return jnp.concatenate(intra_dists_all), jnp.concatenate(inter_dists_all)


def silhouette_samples(X: np.ndarray, labels: np.ndarray, chunk_size: int = 256) -> np.ndarray:
    """Compute the Silhouette Coefficient for each observation.

    Implements :func:`sklearn.metrics.silhouette_samples`.

    Parameters
    ----------
    X
        Array of shape (n_cells, n_features) representing a
        feature array.
    labels
        Array of shape (n_cells,) representing label values
        for each observation.
    chunk_size
        Number of samples to process at a time for distance computation.

    Returns
    -------
    silhouette scores array of shape (n_cells,)
    """
    if X.shape[0] != labels.shape[0]:
        raise ValueError("X and labels should have the same number of samples")
    labels = pd.Categorical(labels).codes
    labels = jnp.asarray(labels)
    label_freqs = jnp.bincount(labels)
    reduce_fn = partial(_silhouette_reduce, labels=labels, label_freqs=label_freqs)
    results = _pairwise_distances_chunked(X, chunk_size=chunk_size, reduce_fn=reduce_fn)
    intra_clust_dists, inter_clust_dists = results

    denom = jnp.take(label_freqs - 1, labels, mode="clip")
    intra_clust_dists /= denom
    sil_samples = inter_clust_dists - intra_clust_dists
    sil_samples /= jnp.maximum(intra_clust_dists, inter_clust_dists)
    return get_ndarray(jnp.nan_to_num(sil_samples))



from functools import partial
from typing import Tuple, Union

import chex
import jax
import jax.numpy as jnp
import numpy as np

from ._utils import get_ndarray

NdArray = Union[np.ndarray, jnp.ndarray]


@chex.dataclass
class _NeighborProbabilityState:
    H: float
    P: chex.ArrayDevice
    Hdiff: float
    beta: float
    betamin: float
    betamax: float
    tries: int


@jax.jit
def _Hbeta(knn_dists_row: jnp.ndarray, beta: float) -> Tuple[jnp.ndarray, jnp.ndarray]:
    P = jnp.exp(-knn_dists_row * beta)
    sumP = jnp.nansum(P)
    H = jnp.where(sumP == 0, 0, jnp.log(sumP) + beta * jnp.nansum(knn_dists_row * P) / sumP)
    P = jnp.where(sumP == 0, jnp.zeros_like(knn_dists_row), P / sumP)
    return H, P


@jax.jit
def _get_neighbor_probability(
    knn_dists_row: jnp.ndarray, perplexity: float, tol: float
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    beta = 1
    betamin = -jnp.inf
    betamax = jnp.inf
    H, P = _Hbeta(knn_dists_row, beta)
    Hdiff = H - jnp.log(perplexity)

    def _get_neighbor_probability_step(state):
        Hdiff = state.Hdiff
        beta = state.beta
        betamin = state.betamin
        betamax = state.betamax
        tries = state.tries

        new_betamin = jnp.where(Hdiff > 0, beta, betamin)
        new_betamax = jnp.where(Hdiff > 0, betamax, beta)
        new_beta = jnp.where(
            Hdiff > 0,
            jnp.where(betamax == jnp.inf, beta * 2, (beta + betamax) / 2),
            jnp.where(betamin == -jnp.inf, beta / 2, (beta + betamin) / 2),
        )
        new_H, new_P = _Hbeta(knn_dists_row, new_beta)
        new_Hdiff = new_H - jnp.log(perplexity)
        return _NeighborProbabilityState(
            H=new_H, P=new_P, Hdiff=new_Hdiff, beta=new_beta, betamin=new_betamin, betamax=new_betamax, tries=tries + 1
        )

    def _get_neighbor_probability_convergence(state):
        Hdiff, tries = state.Hdiff, state.tries
        return jnp.logical_and(jnp.abs(Hdiff) > tol, tries < 50)

    init_state = _NeighborProbabilityState(H=H, P=P, Hdiff=Hdiff, beta=beta, betamin=betamin, betamax=betamax, tries=0)
    final_state = jax.lax.while_loop(_get_neighbor_probability_convergence, _get_neighbor_probability_step, init_state)
    return final_state.H, final_state.P


def _compute_simpson_index_cell(
    knn_dists_row: jnp.ndarray, knn_labels_row: jnp.ndarray, n_batches: int, perplexity: float, tol: float
) -> jnp.ndarray:
    H, P = _get_neighbor_probability(knn_dists_row, perplexity, tol)

    def _non_zero_H_simpson():
        sumP = jnp.bincount(knn_labels_row, weights=P, length=n_batches)
        return jnp.where(knn_labels_row.shape[0] == P.shape[0], jnp.dot(sumP, sumP), 1)

    return jnp.where(H == 0, -1, _non_zero_H_simpson())


def compute_simpson_index(
    knn_dists: NdArray,
    knn_idx: NdArray,
    labels: NdArray,
    n_labels: int,
    perplexity: float = 30,
    tol: float = 1e-5,
) -> np.ndarray:
    """Compute the Simpson index for each cell.

    Parameters
    ----------
    knn_dists
        KNN distances of size (n_cells, n_neighbors).
    knn_idx
        KNN indices of size (n_cells, n_neighbors) corresponding to distances.
    labels
        Cell labels of size (n_cells,).
    n_labels
        Number of labels.
    perplexity
        Measure of the effective number of neighbors.
    tol
        Tolerance for binary search.

    Returns
    -------
    simpson_index
        Simpson index of size (n_cells,).
    """
    knn_dists = jnp.array(knn_dists)
    knn_idx = jnp.array(knn_idx)
    labels = jnp.array(labels)
    knn_labels = labels[knn_idx]
    simpson_fn = partial(_compute_simpson_index_cell, n_batches=n_labels, perplexity=perplexity, tol=tol)
    out = jax.vmap(simpson_fn)(knn_dists, knn_labels)
    return get_ndarray(out)



import functools

import jax
import jax.numpy as jnp
import numpy as np

from scib_metrics.utils import cdist, get_ndarray

from ._dataclass import NeighborsOutput


@functools.partial(jax.jit, static_argnames=["k", "recall_target"])
def _euclidean_ann(qy: jnp.ndarray, db: jnp.ndarray, k: int, recall_target: float = 0.95):
    """Compute half squared L2 distance between query points and database points."""
    dists = cdist(qy, db)
    return jax.lax.approx_min_k(dists, k=k, recall_target=recall_target)


def jax_approx_min_k(
    X: np.ndarray, n_neighbors: int, recall_target: float = 0.95, chunk_size: int = 2048
) -> NeighborsOutput:
    """Run approximate nearest neighbor search using jax.

    On TPU backends, this is approximate nearest neighbor search. On other backends, this is exact nearest neighbor search.

    Parameters
    ----------
    X
        Data matrix.
    n_neighbors
        Number of neighbors to search for.
    recall_target
        Target recall for approximate nearest neighbor search.
    chunk_size
        Number of query points to search for at once.
    """
    db = jnp.asarray(X)
    # Loop over query points in chunks
    neighbors = []
    dists = []
    for i in range(0, db.shape[0], chunk_size):
        start = i
        end = min(i + chunk_size, db.shape[0])
        qy = db[start:end]
        dist, neighbor = _euclidean_ann(qy, db, k=n_neighbors, recall_target=recall_target)
        neighbors.append(neighbor)
        dists.append(dist)
    neighbors = jnp.concatenate(neighbors, axis=0)
    dists = jnp.concatenate(dists, axis=0)
    return NeighborsOutput(indices=get_ndarray(neighbors), distances=get_ndarray(dists))



import os
import warnings
from dataclasses import asdict, dataclass
from enum import Enum
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
from anndata import AnnData
from plottable import ColumnDefinition, Table
from plottable.cmap import normed_cmap
from plottable.plots import bar
from sklearn.preprocessing import MinMaxScaler
from tqdm import tqdm

import scib_metrics
from scib_metrics.nearest_neighbors import NeighborsOutput, pynndescent

Kwargs = Dict[str, Any]
MetricType = Union[bool, Kwargs]

_LABELS = "labels"
_BATCH = "batch"
_X_PRE = "X_pre"
_METRIC_TYPE = "Metric Type"
_AGGREGATE_SCORE = "Aggregate score"

# Mapping of metric fn names to clean DataFrame column names
metric_name_cleaner = {
    "silhouette_label": "Silhouette label",
    "silhouette_batch": "Silhouette batch",
    "isolated_labels": "Isolated labels",
    "nmi_ari_cluster_labels_leiden_nmi": "Leiden NMI",
    "nmi_ari_cluster_labels_leiden_ari": "Leiden ARI",
    "nmi_ari_cluster_labels_kmeans_nmi": "KMeans NMI",
    "nmi_ari_cluster_labels_kmeans_ari": "KMeans ARI",
    "clisi_knn": "cLISI",
    "ilisi_knn": "iLISI",
    "kbet_per_label": "KBET",
    "graph_connectivity": "Graph connectivity",
    "pcr_comparison": "PCR comparison",
}


@dataclass(frozen=True)
class BioConservation:
    """Specification of bio conservation metrics to run in the pipeline.

    Metrics can be included using a boolean flag. Custom keyword args can be
    used by passing a dictionary here. Keyword args should not set data-related
    parameters, such as `X` or `labels`.
    """

    isolated_labels: MetricType = True
    nmi_ari_cluster_labels_leiden: MetricType = True
    nmi_ari_cluster_labels_kmeans: MetricType = False
    silhouette_label: MetricType = True
    clisi_knn: MetricType = True


@dataclass(frozen=True)
class BatchCorrection:
    """Specification of which batch correction metrics to run in the pipeline.

    Metrics can be included using a boolean flag. Custom keyword args can be
    used by passing a dictionary here. Keyword args should not set data-related
    parameters, such as `X` or `labels`.
    """

    silhouette_batch: MetricType = True
    ilisi_knn: MetricType = True
    kbet_per_label: MetricType = True
    graph_connectivity: MetricType = True
    pcr_comparison: MetricType = True


class MetricAnnDataAPI(Enum):
    """Specification of the AnnData API for a metric."""

    isolated_labels = lambda ad, fn: fn(ad.X, ad.obs[_LABELS], ad.obs[_BATCH])
    nmi_ari_cluster_labels_leiden = lambda ad, fn: fn(ad.obsp["15_connectivities"], ad.obs[_LABELS])
    nmi_ari_cluster_labels_kmeans = lambda ad, fn: fn(ad.X, ad.obs[_LABELS])
    silhouette_label = lambda ad, fn: fn(ad.X, ad.obs[_LABELS])
    clisi_knn = lambda ad, fn: fn(ad.obsp["90_distances"], ad.obs[_LABELS])
    graph_connectivity = lambda ad, fn: fn(ad.obsp["15_distances"], ad.obs[_LABELS])
    silhouette_batch = lambda ad, fn: fn(ad.X, ad.obs[_LABELS], ad.obs[_BATCH])
    pcr_comparison = lambda ad, fn: fn(ad.obsm[_X_PRE], ad.X, ad.obs[_BATCH], categorical=True)
    ilisi_knn = lambda ad, fn: fn(ad.obsp["90_distances"], ad.obs[_BATCH])
    kbet_per_label = lambda ad, fn: fn(ad.obsp["50_connectivities"], ad.obs[_BATCH], ad.obs[_LABELS])


class Benchmarker:
    """Benchmarking pipeline for the single-cell integration task.

    Parameters
    ----------
    adata
        AnnData object containing the raw count data and integrated embeddings as obsm keys.
    batch_key
        Key in `adata.obs` that contains the batch information.
    label_key
        Key in `adata.obs` that contains the cell type labels.
    embedding_obsm_keys
        List of obsm keys that contain the embeddings to be benchmarked.
    bio_conservation_metrics
        Specification of which bio conservation metrics to run in the pipeline.
    batch_correction_metrics
        Specification of which batch correction metrics to run in the pipeline.
    pre_integrated_embedding_obsm_key
        Obsm key containing a non-integrated embedding of the data. If `None`, the embedding will be computed
        in the prepare step. See the notes below for more information.
    n_jobs
        Number of jobs to use for parallelization of neighbor search.

    Notes
    -----
    `adata.X` should contain a form of the data that is not integrated, but is normalized. The `prepare` method will
    use `adata.X` for PCA via :func:`~scanpy.tl.pca`, which also only uses features masked via `adata.var['highly_variable']`.

    See further usage examples in the following tutorial:

    1. :doc:`/notebooks/lung_example`
    """

    def __init__(
        self,
        adata: AnnData,
        batch_key: str,
        label_key: str,
        embedding_obsm_keys: List[str],
        bio_conservation_metrics: Optional[BioConservation] = None,
        batch_correction_metrics: Optional[BatchCorrection] = None,
        pre_integrated_embedding_obsm_key: Optional[str] = None,
        n_jobs: int = 1,
    ):
        self._adata = adata
        self._embedding_obsm_keys = embedding_obsm_keys
        self._pre_integrated_embedding_obsm_key = pre_integrated_embedding_obsm_key
        self._bio_conservation_metrics = bio_conservation_metrics if bio_conservation_metrics else BioConservation()
        self._batch_correction_metrics = batch_correction_metrics if batch_correction_metrics else BatchCorrection()
        self._results = pd.DataFrame(columns=list(self._embedding_obsm_keys) + [_METRIC_TYPE])
        self._emb_adatas = {}
        self._neighbor_values = (15, 50, 90)
        self._prepared = False
        self._benchmarked = False
        self._batch_key = batch_key
        self._label_key = label_key
        self._n_jobs = n_jobs

        self._metric_collection_dict = {
            "Bio conservation": self._bio_conservation_metrics,
            "Batch correction": self._batch_correction_metrics,
        }

    def prepare(self, neighbor_computer: Optional[Callable[[np.ndarray, int], NeighborsOutput]] = None) -> None:
        """Prepare the data for benchmarking.

        Parameters
        ----------
        neighbor_computer
            Function that computes the neighbors of the data. If `None`, the neighbors will be computed
            with :func:`~scib_metrics.utils.nearest_neighbors.pynndescent`. The function should take as input
            the data and the number of neighbors to compute and return a :class:`~scib_metrics.utils.nearest_neighbors.NeighborsOutput`
            object.
        """
        # Compute PCA
        if self._pre_integrated_embedding_obsm_key is None:
            # This is how scib does it
            # https://github.com/theislab/scib/blob/896f689e5fe8c57502cb012af06bed1a9b2b61d2/scib/metrics/pcr.py#L197
            sc.tl.pca(self._adata, use_highly_variable=False)
            self._pre_integrated_embedding_obsm_key = "X_pca"

        for emb_key in self._embedding_obsm_keys:
            self._emb_adatas[emb_key] = AnnData(self._adata.obsm[emb_key], obs=self._adata.obs)
            self._emb_adatas[emb_key].obs[_BATCH] = np.asarray(self._adata.obs[self._batch_key].values)
            self._emb_adatas[emb_key].obs[_LABELS] = np.asarray(self._adata.obs[self._label_key].values)
            self._emb_adatas[emb_key].obsm[_X_PRE] = self._adata.obsm[self._pre_integrated_embedding_obsm_key]

        # Compute neighbors
        for ad in tqdm(self._emb_adatas.values(), desc="Computing neighbors"):
            if neighbor_computer is not None:
                neigh_output = neighbor_computer(ad.X, max(self._neighbor_values))
            else:
                neigh_output = pynndescent(
                    ad.X, n_neighbors=max(self._neighbor_values), random_state=0, n_jobs=self._n_jobs
                )
            indices, distances = neigh_output.indices, neigh_output.distances
            for n in self._neighbor_values:
                sp_distances, sp_conns = sc.neighbors._compute_connectivities_umap(
                    indices[:, :n], distances[:, :n], ad.n_obs, n_neighbors=n
                )
                ad.obsp[f"{n}_connectivities"] = sp_conns
                ad.obsp[f"{n}_distances"] = sp_distances

        self._prepared = True

    def benchmark(self) -> None:
        """Run the pipeline."""
        if self._benchmarked:
            warnings.warn(
                "The benchmark has already been run. Running it again will overwrite the previous results.",
                UserWarning,
            )

        if not self._prepared:
            self.prepare()

        num_metrics = sum(
            [sum([v is not False for v in asdict(met_col)]) for met_col in self._metric_collection_dict.values()]
        )

        for emb_key, ad in tqdm(self._emb_adatas.items(), desc="Embeddings", position=0, colour="green"):
            pbar = tqdm(total=num_metrics, desc="Metrics", position=1, leave=False, colour="blue")
            for metric_type, metric_collection in self._metric_collection_dict.items():
                for metric_name, use_metric_or_kwargs in asdict(metric_collection).items():
                    if use_metric_or_kwargs:
                        metric_fn = getattr(scib_metrics, metric_name)
                        if isinstance(use_metric_or_kwargs, dict):
                            # Kwargs in this case
                            metric_fn = partial(metric_fn, **use_metric_or_kwargs)
                        metric_value = getattr(MetricAnnDataAPI, metric_name)(ad, metric_fn)
                        # nmi/ari metrics return a dict
                        if isinstance(metric_value, dict):
                            for k, v in metric_value.items():
                                self._results.loc[f"{metric_name}_{k}", emb_key] = v
                                self._results.loc[f"{metric_name}_{k}", _METRIC_TYPE] = metric_type
                        else:
                            self._results.loc[metric_name, emb_key] = metric_value
                            self._results.loc[metric_name, _METRIC_TYPE] = metric_type
                        pbar.update(1)

        self._benchmarked = True

    def get_results(self, min_max_scale: bool = True, clean_names: bool = True) -> pd.DataFrame:
        """Return the benchmarking results.

        Parameters
        ----------
        min_max_scale
            Whether to min max scale the results.
        clean_names
            Whether to clean the metric names.

        Returns
        -------
        The benchmarking results.
        """
        df = self._results.transpose()
        df.index.name = "Embedding"
        df = df.loc[df.index != _METRIC_TYPE]
        if min_max_scale:
            # Use sklearn to min max scale
            df = pd.DataFrame(
                MinMaxScaler().fit_transform(df),
                columns=df.columns,
                index=df.index,
            )
        if clean_names:
            df = df.rename(columns=metric_name_cleaner)
        df = df.transpose()
        df[_METRIC_TYPE] = self._results[_METRIC_TYPE].values

        # Compute scores
        per_class_score = df.groupby(_METRIC_TYPE).mean().transpose()
        # This is the default scIB weighting from the manuscript
        per_class_score["Total"] = 0.4 * per_class_score["Batch correction"] + 0.6 * per_class_score["Bio conservation"]
        df = pd.concat([df.transpose(), per_class_score], axis=1)
        df.loc[_METRIC_TYPE, per_class_score.columns] = _AGGREGATE_SCORE
        return df

    def plot_results_table(
        self, min_max_scale: bool = True, show: bool = True, save_dir: Optional[str] = None
    ) -> Table:
        """Plot the benchmarking results.

        Parameters
        ----------
        min_max_scale
            Whether to min max scale the results.
        show
            Whether to show the plot.
        save_dir
            The directory to save the plot to. If `None`, the plot is not saved.
        """
        num_embeds = len(self._embedding_obsm_keys)
        cmap_fn = lambda col_data: normed_cmap(col_data, cmap=matplotlib.cm.PRGn, num_stds=2.5)
        df = self.get_results(min_max_scale=min_max_scale)
        # Do not want to plot what kind of metric it is
        plot_df = df.drop(_METRIC_TYPE, axis=0)
        # Sort by total score
        plot_df = plot_df.sort_values(by="Total", ascending=False).astype(np.float64)
        plot_df["Method"] = plot_df.index

        # Split columns by metric type, using df as it doesn't have the new method col
        score_cols = df.columns[df.loc[_METRIC_TYPE] == _AGGREGATE_SCORE]
        other_cols = df.columns[df.loc[_METRIC_TYPE] != _AGGREGATE_SCORE]
        column_definitions = [
            ColumnDefinition("Method", width=1.5, textprops={"ha": "left", "weight": "bold"}),
        ]
        # Circles for the metric values
        column_definitions += [
            ColumnDefinition(
                col,
                title=col.replace(" ", "\n", 1),
                width=1,
                textprops={
                    "ha": "center",
                    "bbox": {"boxstyle": "circle", "pad": 0.25},
                },
                cmap=cmap_fn(plot_df[col]),
                group=df.loc[_METRIC_TYPE, col],
                formatter="{:.2f}",
            )
            for i, col in enumerate(other_cols)
        ]
        # Bars for the aggregate scores
        column_definitions += [
            ColumnDefinition(
                col,
                width=1,
                title=col.replace(" ", "\n", 1),
                plot_fn=bar,
                plot_kw={
                    "cmap": matplotlib.cm.YlGnBu,
                    "plot_bg_bar": False,
                    "annotate": True,
                    "height": 0.9,
                    "formatter": "{:.2f}",
                },
                group=df.loc[_METRIC_TYPE, col],
                border="left" if i == 0 else None,
            )
            for i, col in enumerate(score_cols)
        ]
        # Allow to manipulate text post-hoc (in illustrator)
        with matplotlib.rc_context({"svg.fonttype": "none"}):
            fig, ax = plt.subplots(figsize=(len(df.columns) * 1.25, 3 + 0.3 * num_embeds))
            tab = Table(
                plot_df,
                cell_kw={
                    "linewidth": 0,
                    "edgecolor": "k",
                },
                column_definitions=column_definitions,
                ax=ax,
                row_dividers=True,
                footer_divider=True,
                textprops={"fontsize": 10, "ha": "center"},
                row_divider_kw={"linewidth": 1, "linestyle": (0, (1, 5))},
                col_label_divider_kw={"linewidth": 1, "linestyle": "-"},
                column_border_kw={"linewidth": 1, "linestyle": "-"},
                index_col="Method",
            ).autoset_fontcolors(colnames=plot_df.columns)
        if show:
            plt.show()
        if save_dir is not None:
            fig.savefig(os.path.join(save_dir, "scib_results.svg"), facecolor=ax.get_facecolor(), dpi=300)

        return tab
